from multiprocessing import Queue
from threading import Thread
import threading
import traceback
from typing import TypeVar, Generic, List, Optional

from tudelft.utilities.listener.Listener import Listener
from uri.uri import URI

from geniusweb.connection.ConnectionEnd import ConnectionEnd
from geniusweb.references.Reference import Reference


S=TypeVar("S")

class Info (Generic[S]):
	'''
	Wrapper around data so that we can put Null end EOS in a
	{@link ArrayBlockingQueue}
	@param <S> the type of contained data.
	'''

class Data (Info[S]):
	data:S

	def __init__(self, data:S):
		self._data = data

	def get(self)->S:
		return self._data;

class EOS (Info[S]):
	'''
	End of stream.
	
	@author wouter
	'''
	
# I use this single instance everywhere
THE_EOS:EOS = EOS()


IN = TypeVar('IN')
OUT = TypeVar('OUT')

class BasicConnection(ConnectionEnd[IN, OUT]):
	'''
	A basic connection that implements connection with direct calls
	
	@param <IN>  the type of the incoming data
	@param <OUT> the type of outgoing data
	'''
	_listeners:List[Listener[IN]] = []
	_synclock = threading.RLock()
	_error:Exception

	# to be initialized
	_handler:Optional[Listener[OUT]] = None
	_messages:Queue = Queue(maxsize=4) # mypy needs type annotation??

	def __init__(self, reference:Reference , uri:URI ):
		'''
		@param reference Reference that was used to create this connection.
		@param uri       the URI of the remote endpoint that makes up the
		                 connection. This is a URI that uniquely identifies the
		                 remote object
		'''
		self._reference = reference
		self._uri = uri

	def init(self, newhandler:Listener[OUT] ) :
		'''
		To be called to hook up the other side that will handle a send action
		from us. Must be called first.
		
		@param newhandler a Listener&lt;OUT&gt; that can handle send actions.
		'''
		if self._handler: 
			raise ValueError("already initialized")
		self._handler = newhandler
		this=self

		class MyHandlerThread(Thread):
			def run(self, messages):
				try:
					while (True):
						mess = this._messages.take()
						if mess==THE_EOS:
							break;
						this._handler.notifyChange(mess.get())
				except Exception as e:
					this.setError(e)
				this._handler = None
				print("BasicConnection closed");
		
		handlerThread=MyHandlerThread()
		handlerThread.start();

	def setError(self,  e:Exception):
		'''
		Error condition occurs. Record error and close connection
		
		@param e
		'''
		with self._synclock:
			if not self._error:
				# maybe log instead?
				traceback.print_exc()
				self._error = e
				self.close()

	def send(self, data:OUT ) :
		if not self._handler:
			raise ValueError(
				"BasicConnection has not been initialized or was closed")
		# it seems there is no InterruptedException possible in python.
		self._messages.put(Data(data))

	def getReference(self) -> Reference :
		return self._reference

	def getRemoteURI(self)->URI: 
		return self._uri

	def close(self):
		print("flushing and terminating " + str(self))
		if not self._handler or THE_EOS in self._messages:
			return
		# it seems there is no InterruptedException possible in python.
		self._messages.put(self.THE_EOS)

	def __repr__(self):
		return "BasicConnection[" + str(self._reference) + "]"

	def getError(self)->Exception:
		return self._error;

	def isOpen(self)->bool: 
		'''
		 @return true iff this connection is open
		'''
		return self._handler != None

	#****************** implements listenable ****************
	# override because notifyListeners should throw exceptions.
	def addListener(self, l:Listener[IN]):
		self._listeners.append(l)

	def removeListener(self, l:Listener[IN] ) :
		self._listeners.remove(l)

	def notifyListeners(self, data:IN ) :
		for l in self._listeners:
			l.notifyChange(data)

